#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

from typing import List

from awstt.worker.scanner import Scanner
from awstt.worker.types import AWSResource, AWSResourceTag


@Scanner.register("IAM:User")
class IAMUserScanner(Scanner):
    def __init__(self, partition, _, credential):
        if partition in ["aws", "aws-gov"]:
            super().__init__(partition, ["global"], credential)
        else:
            super().__init__(partition, ["cn-northwest-1", "cn-northeast-1"], credential)

    def build_resource(self, client: any, user: dict) -> AWSResource:
        return AWSResource(
            self.category,
            self._build_arn(client, user["Arn"]),
            [AWSResourceTag(tag["Key"], tag["Value"]) for tag in user.get("Tags", [])],
            user,
        )

    def _list_resources(self, client: any) -> List[AWSResource]:
        resources = []
        paginator = client.get_paginator("list_users").paginate()

        for page in paginator:
            for user in page.get("Users", []):
                resources.append(self.build_resource(client, user))

        return resources

    @property
    def _service_name(self) -> str:
        return "iam"

    @property
    def _arn_resource_type(self) -> str:
        return "user"

    @property
    def category(self) -> str:
        return "IAM:User"



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

import importlib
import inspect
from collections import defaultdict
from collections.abc import ItemsView
from typing import Callable, Dict, List, Optional, Type, TypeVar
from warnings import warn


T = TypeVar("T")
"""
Subclass type.
"""

HookType = Callable[[Type[T], str], None]
"""
Hooks are callables that take the subclass and registered name of the subclass.
"""


class Error(Exception):
    pass


class RegistrationError(Error):
    pass


class Registrable:
    """
    Adapted from `allennlp.common.registrable
    <https://github.com/allenai/allennlp/blob/master/allennlp/common/registrable.py>`_.

    Any class that inherits from ``Registrable`` gains access to a named registry for its
    subclasses. To register them, just decorate them with the classmethod
    ``@BaseClass.register(name)``.

    After which you can call ``BaseClass.list_available()`` to get the keys for the
    registered subclasses, and ``BaseClass.by_name(name)`` to get the corresponding subclass.
    Note that the registry stores the subclasses themselves; not class instances.

    You can specify a default by setting ``BaseClass.default_implementation``.
    If it is set, it will be the first element of ``list_available()``.

    Note that if you use this class to implement a new ``Registrable`` abstract class,
    you must ensure that all subclasses of the abstract class are loaded when the module is
    loaded, because the subclasses register themselves in their respective files. You can
    achieve this by having the abstract class and all subclasses in the ``__init__.py`` of the
    module in which they reside (as this causes any import of either the abstract class or
    a subclass to load all other subclasses and the abstract class).
    """

    _registry: Dict[Type, Dict[str, Type]] = defaultdict(dict)
    _hooks: Optional[List[HookType]] = None

    default_implementation: Optional[str] = None
    """
    Optional name of default implementation. If specified, the default will be listed
    first in :func:`registrable.Registrable.list_available`.
    """

    @classmethod
    def register(
        cls: Type[T],
        name: str,
        override: bool = False,
        hooks: Optional[List[HookType]] = None,
    ):
        """
        Class decorator for registering a subclass.

        Parameters
        ----------
        name : ``str``
            The name to register the subclass under.
        override : ``bool``, optional (default = False)
            If ``name`` is already registered a :class:`registrable.exceptions.RegistrationError` will be raised
            unless this is set to ``True``.
        hooks : ``Optional[List[HookType]]``, optional (default = None)
            Hooks to run when the subclass is registered.

        Raises
        ------
        RegistrationError
        """
        registry = Registrable._registry[cls]
        default_hooks = cls._hooks or []  # type: ignore

        def add_subclass_to_registry(subclass: Type[T]):
            if not inspect.isclass(subclass) or not issubclass(subclass, cls):
                raise RegistrationError(
                    f"Cannot register {subclass.__name__} as {name}; "
                    f"{subclass.__name__} must be a subclass of {cls.__name__}"
                )
            # Add to registry.
            # If name already registered, warn if overriding or raise an error if override not allowed.
            if name in registry:
                if not override:
                    raise RegistrationError(
                        f"Cannot register {subclass.__name__} as {name}; "
                        f"name already in use for {registry[name].__name__}"
                    )
                else:
                    warn(f"Overriding {name} in {cls.__name__} registry")
            registry[name] = subclass
            for hook in default_hooks + (hooks or []):
                hook(subclass, name)
            return subclass

        return add_subclass_to_registry

    @classmethod
    def hook(cls, hook: HookType):
        """
        Function decorator for adding a default hook to a registrable base class.
        """
        if not cls._hooks:
            cls._hooks = []
        cls._hooks.append(hook)
        return hook

    @classmethod
    def by_name(cls: Type[T], name: str) -> type | Type[T]:
        """
        Get a subclass by its registered name, or its fully qualified class name.

        Parameters
        ----------
        name : ``str``

        Returns
        -------
        Type[T]
            The subclass registered under ``name``.

        Raises
        ------
        RegistrationError
            If name is not a registered subclass or valid fully qualified class name.
        """
        if name in Registrable._registry[cls]:
            return Registrable._registry[cls][name]
        elif "." in name:
            # This might be a fully qualified class name, so we'll try importing its "module"
            # and finding it there.
            parts = name.split(".")
            submodule = ".".join(parts[:-1])
            class_name = parts[-1]

            try:
                module = importlib.import_module(submodule)
            except ModuleNotFoundError:
                raise RegistrationError(
                    f"tried to interpret {name} as a path to a class " f"but unable to import module {submodule}"
                )

            try:
                maybe_subclass = getattr(module, class_name)
            except AttributeError:
                raise RegistrationError(
                    f"tried to interpret {name} as a path to a class "
                    f"but unable to find class {class_name} in {submodule}"
                )

            if not inspect.isclass(maybe_subclass) or not issubclass(maybe_subclass, cls):
                raise RegistrationError(
                    f"tried to interpret {name} as a path to a class "
                    f"but {class_name} is not a subclass of {cls.__name__}"
                )

            # Add subclass to registry and return it.
            Registrable._registry[cls][name] = maybe_subclass
            return maybe_subclass
        else:
            # is not a qualified class name
            raise RegistrationError(f"{name} is not a registered name for {cls.__name__}.")

    @classmethod
    def list_available(cls: Type[T], in_order: bool = True) -> List[str]:
        """
        List all registered subclasses.

        If ``cls.default_implementation`` is specified, it will be first in the list.
        """
        keys = list(Registrable._registry[cls].keys())
        if in_order:
            keys.sort()

        default = cls.default_implementation  # type: ignore

        if default is None:
            return keys
        if default not in keys:
            raise RegistrationError(f"Default implementation {default} is not registered")
        return [default] + [k for k in keys if k != default]

    @classmethod
    def is_registered(cls: Type[T], name: str) -> bool:
        """
        Returns True if ``name`` is a registered name.
        """
        return name in Registrable._registry[cls]

    @classmethod
    def iter_registered(cls: Type[T]) -> ItemsView[str, type]:
        """
        Iterate through the registered names and subclasses.
        """
        return Registrable._registry[cls].items()



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

from awstt.worker.scanners.cloudformation import CloudFormationScanner
from awstt.worker.scanners.cloudfront_distribution import CloudFrontScanner
from awstt.worker.scanners.cloudtrail import CloudTrailScanner
from awstt.worker.scanners.cloudwatch_alarm import CloudWatchAlarmScanner
from awstt.worker.scanners.cloudwatch_loggroup import CloudWatchLogGroupScanner
from awstt.worker.scanners.cloudwatch_rules import CloudWatchRulesScanner
from awstt.worker.scanners.dynamodb_table import DynamoDBScanner
from awstt.worker.scanners.ec2_ami import AMIScanner
from awstt.worker.scanners.ec2_autoscaling import AutoScalingGroupScanner
from awstt.worker.scanners.ec2_ebs import EBSScanner
from awstt.worker.scanners.ec2_eip import EIPScanner
from awstt.worker.scanners.ec2_instance import EC2Scanner
from awstt.worker.scanners.ec2_internet_gateway import InternetGatewayScanner
from awstt.worker.scanners.ec2_nat_gateway import NATGatewayScanner
from awstt.worker.scanners.ec2_network_acl import NetworkACLScanner
from awstt.worker.scanners.ec2_routetable import RouteTableScanner
from awstt.worker.scanners.ec2_securitygroup import SecurityGroupScanner
from awstt.worker.scanners.ec2_snapshot import SnapshotScanner
from awstt.worker.scanners.ec2_vpc import VPCScanner
from awstt.worker.scanners.ec2_vpc_peering import VPCPeeringScanner
from awstt.worker.scanners.ec2_vpn_gateway import VPNGatewayScanner
from awstt.worker.scanners.efs import EFSScanner
from awstt.worker.scanners.elasticache_cluster import ElastiCacheClusterScanner
from awstt.worker.scanners.elasticache_snapshot import ElastiCacheSnapshotScanner
from awstt.worker.scanners.elasticsearch import ElasticsearchScanner
from awstt.worker.scanners.elb import ELBScanner
from awstt.worker.scanners.emr_cluster import EMRClusterScanner
from awstt.worker.scanners.emr_studio import EMRStudioScanner
from awstt.worker.scanners.iam_policy import IAMPolicyScanner
from awstt.worker.scanners.iam_role import IAMRoleScanner
from awstt.worker.scanners.iam_user import IAMUserScanner
from awstt.worker.scanners.kms import KMSScanner
from awstt.worker.scanners.lambda_function import LambdaFunctionScanner
from awstt.worker.scanners.lightsail_instance import LightsailScanner
from awstt.worker.scanners.opensearch import OpenSearchScanner
from awstt.worker.scanners.rds_cluster import RDSClusterScanner
from awstt.worker.scanners.rds_instance import RDSInstanceScanner
from awstt.worker.scanners.s3_bucket import S3BucketScanner



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

import logging

import executor
from config import init_config


logger = logging.getLogger(__name__)


def execute(evt: any, _):
    """
    Lambda input uses dict in format of json configure.
    """

    if evt.get("action", None) is None:
        return

    config = init_config(
        dict(
            action=evt.get("action", None),
            tags=evt.get("tags", []),
            filters=evt.get("filters", []),
            resources=evt.get("resources", []),
            force=evt.get("force", False),
        )
    )

    executor.run(config)



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

import json
import logging
import os
from typing import Optional

import rich_click as click

from awstt import executor
from awstt.config import init_config
from awstt.output import init_logger


logger = logging.getLogger()

click.rich_click.USE_RICH_MARKUP = False

click.rich_click.COMMAND_GROUPS = {
    "awstt": [{"name": "Commands", "commands": ["set", "unset", "list", "exec", "info"]}]
}

_common_option_groups = [
    {
        "name": "Credential Options",
        "options": ["--access_key", "--secret_key", "--profile"],
    },
    {
        "name": "Log Options",
        "options": ["--save-log", "--log-level"],
    },
    {
        "name": "Help",
        "options": ["--help"],
    },
]

click.rich_click.OPTION_GROUPS = {
    "awstt": [
        {
            "name": "Help",
            "options": ["--help"],
        },
    ],
    "awstt info": [
        {
            "name": "Help",
            "options": ["--help"],
        },
    ],
    "awstt set": [
        {
            "name": "Tagging Options",
            "options": ["--tag", "--resource", "--region", "--partition", "--force", "--filter"],
        },
    ],
    "awstt unset": [
        {
            "name": "UnTagging Options",
            "options": ["--tag", "--resource", "--region", "--partition", "--filter"],
        },
    ],
    "awstt list": [
        {
            "name": "List Options",
            "options": ["--resource", "--region", "--partition", "--force", "--filter"],
        },
    ],
    "awstt exec": [
        {
            "name": "Execute Options",
            "options": ["--config", "--filter"],
        },
    ],
}
click.rich_click.OPTION_GROUPS["awstt list"] += _common_option_groups
click.rich_click.OPTION_GROUPS["awstt set"] += _common_option_groups
click.rich_click.OPTION_GROUPS["awstt unset"] += _common_option_groups
click.rich_click.OPTION_GROUPS["awstt exec"] += _common_option_groups


@click.group()
def info():
    """Show helpful information about AWS-Tag-Tools"""
    pass


@info.command("resources", short_help="List all supported resources", add_help_option=False)
def info_resources():
    from awstt.output import console
    from awstt.worker.scanner import Scanner

    console = console()
    table = console.new_table(title="", show_header=False, show_edge=False)
    table.add_column("")
    for resource in Scanner.list_available():
        table.add_row(resource)

    console.print(
        console.new_panel(
            table,
            padding=(1, 1),
            title="[b red]Supported resources",
            border_style="bright_blue",
        ),
    )


_cli_option_partition_choices = click.Choice(["aws", "aws-cn", "aws-us-gov"])

_cli_common_options = [
    click.option("--access_key", help="Access key of AWS AK/SK", default=os.environ.get("AWS_ACCESS_KEY_ID")),
    click.option("--secret_key", help="Secret key of AWS AK/SK", default=os.environ.get("AWS_SECRET_ACCESS_KEY")),
    click.option("--profile", help="Profile name of AWS CLI Credential", default=os.environ.get("AWS_PROFILE")),
    click.option("--save-log", "save_log", is_flag=True, default=False, help="Save log to file"),
    click.option(
        "--log-level",
        "log_level",
        help="Set log level",
        type=click.Choice(["debug", "info", "warning", "error", "critical"], case_sensitive=False),
        default="error",
        show_default=True,
    ),
]


def cmd_common_options(func):
    for option in reversed(_cli_common_options):
        func = option(func)
    return func


@click.group()
def cmd():
    """AWS-Tag-Tools: The missing tag manager for AWS resources"""
    pass


@cmd.command("set", help="Set tag(s) to resources")
@cmd_common_options
@click.option("--tag", required=True, help="Tag to set", metavar="KEY=VALUE[,KEY=VALUE,...]")
@click.option("--region", help="Region of AWS", metavar="REGION1[,REGION2,...]")
@click.option("--partition", help="Partition of AWS", type=_cli_option_partition_choices, default="aws")
@click.option("--resource", help="Resource type or ARN pattern", metavar="RESOURCE1[,RESOURCE2,...]")
@click.option("--filter", "filters", help="JMESPath expression to filter resources")
@click.option("--force", is_flag=True, default=False, help="Force overwrite if tag exists")
def cmd_set(
    tag: str,
    resource: Optional[str],
    region: Optional[str],
    partition: Optional[str],
    filters: Optional[str],
    force: Optional[bool],
    access_key: Optional[str],
    secret_key: Optional[str],
    profile: Optional[str],
    save_log: Optional[bool],
    log_level: Optional[str],
):
    if len(tag) < 1:
        click.echo("Error: must set at least one tag, see the usage of SET command.")
        click.echo(click.get_current_context().get_help())
        return

    init_logger(log_level, save_log)

    opt_tags = tag.split(",")
    tags = []
    for t in opt_tags:
        if "=" in t:
            kv = t.split("=", 1)
            tags.append({"key": kv[0], "value": kv[1]})
        else:
            click.echo(f"Error: invalid tag option")
            return

    inputs = {
        "action": "set",
        "partition": partition,
        "regions": region.split(",") if region else [],
        "filter": filters,
        "tags": tags,
        "resources": resource.split(",") if resource else [],
        "force": force,
        "credential": {
            "access_key": access_key,
            "secret_key": secret_key,
            "profile": profile,
        },
    }

    config = init_config(inputs)
    executor.run(config)


@cmd.command("unset", help="Unset tag(s) from resources")
@cmd_common_options
@click.option("--tag", required=True, help="tag to unset", metavar="KEY1[,KEY2,...]")
@click.option("--region", help="Region of AWS", metavar="REGION1[,REGION2,...]")
@click.option("--partition", help="Partition of AWS", type=_cli_option_partition_choices, default="aws")
@click.option("--resource", help="Resource type or ARN pattern", metavar="RESOURCE1[,RESOURCE2,...]")
@click.option("--filter", "filters", help="JMESPath expression to filter resources")
def cmd_unset(
    tag: str,
    resource: Optional[str],
    region: Optional[str],
    partition: Optional[str],
    filters: Optional[str],
    access_key: Optional[str],
    secret_key: Optional[str],
    profile: Optional[str],
    save_log: Optional[bool],
    log_level: Optional[str],
):
    if len(tag) < 1:
        click.echo("Error: must set at least one tag, see the usage of UNSET command.")
        click.echo(click.get_current_context().get_help())
        return

    init_logger(log_level, save_log)

    inputs = {
        "action": "unset",
        "partition": partition,
        "regions": region.split(",") if region else [],
        "filter": filters,
        "tags": tag.split(","),
        "resources": resource.split(",") if resource else [],
        "credential": {
            "access_key": access_key,
            "secret_key": secret_key,
            "profile": profile,
        },
    }

    config = init_config(inputs)
    executor.run(config)


@cmd.command("list", help="List resources by tag(s) or spec condition(s)")
@cmd_common_options
@click.option("--region", help="Region of AWS", metavar="REGION1[,REGION2,...]")
@click.option("--partition", help="Partition of AWS", type=_cli_option_partition_choices, default="aws")
@click.option("--resource", help="Resource type or ARN pattern", metavar="RESOURCE1[,RESOURCE2,...]")
@click.option("--filter", "filters", help="JMESPath expression to filter resources")
def cmd_list(
    resource: Optional[str],
    region: Optional[str],
    partition: Optional[str],
    filters: Optional[str],
    access_key: Optional[str],
    secret_key: Optional[str],
    profile: Optional[str],
    save_log: Optional[bool],
    log_level: Optional[str],
):
    init_logger(log_level, save_log)

    inputs = {
        "action": "list",
        "partition": partition,
        "regions": region.split(",") if region else [],
        "filter": filters,
        "tags": [],
        "resources": resource.split(",") if resource else [],
        "force": False,
        "credential": {
            "access_key": access_key,
            "secret_key": secret_key,
            "profile": profile,
        },
    }

    config = init_config(inputs)
    executor.run(config)


@cmd.command("exec", help="Execute action with config file")
@cmd_common_options
@click.option("--config", "cfg", help="Config file", type=click.Path(exists=True))
def cmd_exec(
    cfg: str,
    access_key: Optional[str],
    secret_key: Optional[str],
    profile: Optional[str],
    save_log: Optional[bool],
    log_level: Optional[str],
):
    init_logger(log_level, save_log)

    with open(cfg, "r") as f:
        inputs = json.loads(f.read())

        if not inputs.get("credential"):
            inputs["credential"] = {}

        if access_key:
            inputs["credential"]["access_key"] = access_key
        if secret_key:
            inputs["credential"]["secret_key"] = secret_key
        if profile:
            inputs["credential"]["profile"] = profile

        config = init_config(inputs)
        executor.run(config)


def run():
    cmd.add_command(info)
    cmd(prog_name="awstt", auto_envvar_prefix="AWSTT")



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

from typing import List

from awstt.worker.scanner import Scanner
from awstt.worker.types import AWSResource, AWSResourceTag


@Scanner.register("EC2:AutoScalingGroup")
class AutoScalingGroupScanner(Scanner):
    def build_resource(self, client: any, group: dict) -> AWSResource:
        return AWSResource(
            self.category,
            self._build_arn(client, group["AutoScalingGroupARN"]),
            [AWSResourceTag(tag["Key"], tag["Value"]) for tag in group.get("Tags", [])],
            group,
        )

    def _list_resources(self, client: any) -> List[AWSResource]:
        resources = []
        paginator = client.get_paginator("describe_auto_scaling_groups").paginate()

        for page in paginator:
            for group in page.get("AutoScalingGroups", []):
                resources.append(self.build_resource(client, group))

        return resources

    @property
    def _service_name(self) -> str:
        return "autoscaling"

    @property
    def _arn_resource_type(self) -> str:
        return "autoscaling"

    @property
    def category(self) -> str:
        return "EC2:AutoScalingGroup"



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

from typing import List

from awstt.worker.scanner import Scanner
from awstt.worker.types import AWSResource, AWSResourceTag


@Scanner.register("ElastiCache:Snapshot")
class ElastiCacheSnapshotScanner(Scanner):
    def build_resource(self, client: any, snapshot: dict) -> AWSResource:
        arn = snapshot["ARN"]
        resource_tags = client.list_tags_for_resource(ResourceName=arn)

        return AWSResource(
            self.category,
            self._build_arn(client, arn),
            [AWSResourceTag(tag["Key"], tag["Value"]) for tag in resource_tags.get("TagList", [])],
            snapshot,
        )

    def _list_resources(self, client: any) -> List[AWSResource]:
        resources = []
        paginator = client.get_paginator("describe_snapshots").paginate()

        for page in paginator:
            for snapshot in page.get("Snapshots", []):
                resources.append(self.build_resource(client, snapshot))

        return resources

    @property
    def _service_name(self) -> str:
        return "elasticache"

    @property
    def _arn_resource_type(self) -> str:
        return "snapshot"

    @property
    def category(self) -> str:
        return "ElastiCache:Snapshot"



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

import json
import logging
import os
import re
from dataclasses import asdict, dataclass, field
from typing import List, Optional, Union

from awstt.evals import eval_expression


logger = logging.getLogger(__name__)


@dataclass(frozen=True, init=True, repr=True, order=True)
class _Base:
    def dict(self):
        return asdict(self)


@dataclass(frozen=True, init=True, repr=True, order=True)
class Tag(_Base):
    key: str
    value: Optional[str] = ""


@dataclass(frozen=True, init=True, repr=True, order=True)
class Resource(_Base):
    target: str
    tags: List[Union[Tag, str]] = field(default_factory=list)
    filter: Optional[str] = None
    force: Optional[bool] = False

    def __post_init__(self):
        object.__setattr__(
            self, "tags", [Tag(**tag) if isinstance(tag, dict) else tag for tag in self.tags] if self.tags else []
        )


@dataclass(frozen=True, init=True, repr=True, order=True)
class Credential(_Base):
    access_key: Optional[str] = None
    secret_key: Optional[str] = None
    profile: Optional[str] = None


@dataclass(frozen=True, init=True, repr=True, order=True)
class Config(_Base):
    action: str
    force: bool = False
    filter: Optional[str] = None
    partition: Optional[str] = "aws"
    regions: Optional[List[str]] = field(default_factory=list)
    tags: List[Union[Tag, str]] = field(default_factory=list)
    resources: List[Union[str, Resource]] = field(default_factory=list)
    credential: Optional[Credential] = field(default_factory=Credential)
    env: Optional[any] = field(default_factory=dict)


class ConfigError(Exception):
    pass


def init_config(data: dict) -> Config:
    env = os.environ
    env.pop("AWS_ACCESS_KEY_ID", None)
    env.pop("AWS_SECRET_ACCESS_KEY", None)
    env.pop("AWS_SESSION_TOKEN", None)
    env.pop("AWS_DEFAULT_REGION", None)
    env.pop("AWS_REGION", None)
    env.pop("AWS_PROFILE", None)

    data_str = json.dumps(data)
    for exp in re.findall(r"\${(.+?)}\$", data_str):
        exp_value = eval_expression(exp, env)
        data_str = data_str.replace(f"${{{exp}}}$", str(exp_value))

    data = json.loads(data_str)

    data["action"] = data["action"].lower()
    data["partition"] = data["partition"].lower()
    data["regions"] = [r.lower() for r in data["regions"]]

    data["credential"] = Credential(**data["credential"])
    data["resources"] = (
        [Resource(**res) if isinstance(res, dict) else res for res in data["resources"]] if data["resources"] else []
    )
    data["tags"] = [Tag(**tag) if isinstance(tag, dict) else tag for tag in data["tags"]] if data["tags"] else []
    data["env"] = {k: v for k, v in env.items() if k.lower().startswith("awstt_")} if env else {}

    return Config(**data)


def check_config(config: Config):
    if config.credential.profile is not None and (
        config.credential.access_key is not None or config.credential.secret_key is not None
    ):
        raise ConfigError("Only one of credential.profile or credential.access_key/secret_key should be provided")

    if config.credential.profile is None and (
        config.credential.access_key is None or config.credential.secret_key is None
    ):
        raise ConfigError("Either credential.profile or both credential.access_key/secret_key should be provided")

    if config.action == "set":
        if any(x for x in config.tags if not isinstance(x, Tag)):
            raise ConfigError(f"Tags should be a list of Tag")
        for res in [r for r in config.resources if isinstance(r, Resource)]:
            if any(x for x in res.tags if not isinstance(x, Tag)):
                print([type(x) for x in res.tags if not isinstance(x, Tag)])
                raise ConfigError(f"Resource Tags should be a list of Tag - {res}")

    if config.action == "unset":
        if any(x for x in config.tags if isinstance(x, Tag)):
            raise ConfigError(f"Tags should be a list of key str or key selector in JMESPath expression")
        for res in [r for r in config.resources if isinstance(r, Resource)]:
            if any(x for x in res.tags if isinstance(x, Tag)):
                raise ConfigError(
                    f"Resource Tags should be a list of key str or key selector in JMESPath expression - {res}"
                )



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

from typing import List

from awstt.worker.scanner import Scanner
from awstt.worker.types import AWSResource, AWSResourceTag


@Scanner.register("ELB")
class ELBScanner(Scanner):
    def build_resource(self, client: any, lb: dict) -> AWSResource:
        arn = lb.get("LoadBalancerArn")
        resource_tags = []
        for item in client.describe_tags(ResourceArns=[arn]).get("TagDescriptions", []):
            resource_tags.extend(item.get("Tags", []))

        return AWSResource(
            self.category,
            self._build_arn(client, lb["InstanceId"]),
            [AWSResourceTag(tag["Key"], tag["Value"]) for tag in resource_tags],
            lb,
        )

    def _list_resources(self, client: any) -> List[AWSResource]:
        resources = []
        paginator = client.get_paginator("describe_load_balancers").paginate()

        for page in paginator:
            for lb in page.get("LoadBalancers", []):
                resources.append(self.build_resource(client, lb))

        return resources

    @property
    def _service_name(self) -> str:
        return "elbv2"

    @property
    def _arn_service_type(self) -> str:
        return "elasticloadbalancing"

    @property
    def _arn_resource_type(self) -> str:
        return "loadbalancer"

    @property
    def category(self) -> str:
        return "ELB"



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

from typing import List

from awstt.worker.scanner import Scanner
from awstt.worker.types import AWSResource, AWSResourceTag


@Scanner.register("CloudWatch:Alarm")
class CloudWatchAlarmScanner(Scanner):
    def build_resource(self, client: any, alarm: dict) -> AWSResource:
        arn = alarm["AlarmArn"]
        resource_tags = client.list_tags_for_resource(ResourceARN=arn)

        return AWSResource(
            self.category,
            self._build_arn(client, arn),
            [AWSResourceTag(tag["Key"], tag["Value"]) for tag in resource_tags.get("Tags", [])],
            alarm,
        )

    def _list_resources(self, client: any) -> List[AWSResource]:
        resources = []
        paginator = client.get_paginator("describe_alarms").paginate()

        for page in paginator:
            alarms = []
            alarms.extend(page.get("CompositeAlarms", []))
            alarms.extend(page.get("MetricAlarms", []))

            for alarm in alarms:
                resources.append(self.build_resource(client, alarm))

        return resources

    @property
    def _service_name(self) -> str:
        return "cloudwatch"

    @property
    def _arn_resource_type(self) -> str:
        return "alarm"

    @property
    def category(self) -> str:
        return "CloudWatch:Alarm"



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

from typing import List

from awstt.worker.scanner import Scanner
from awstt.worker.types import AWSResource, AWSResourceTag


@Scanner.register("EMR:Cluster")
class EMRClusterScanner(Scanner):
    def build_resource(self, client: any, cluster: dict) -> AWSResource:
        detail = client.describe_cluster(ClusterId=cluster.get("Id")).get("Cluster", {})

        return AWSResource(
            self.category,
            self._build_arn(client, detail.get("ClusterArn")),
            [AWSResourceTag(tag["Key"], tag["Value"]) for tag in detail.get("Tags", [])],
            detail,
        )

    def _list_resources(self, client: any) -> List[AWSResource]:
        resources = []
        paginator = client.get_paginator("list_clusters").paginate()

        for page in paginator:
            for cluster in page.get("Clusters", []):
                resources.append(self.build_resource(client, cluster))

        return resources

    @property
    def _service_name(self) -> str:
        return "emr"

    @property
    def _arn_resource_type(self) -> str:
        return "cluster"

    @property
    def category(self) -> str:
        return "EMR:Cluster"



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

from typing import List

from awstt.worker.scanner import Scanner
from awstt.worker.types import AWSResource, AWSResourceTag


@Scanner.register("CloudFront:Distribution")
class CloudFrontScanner(Scanner):
    def __init__(self, partition, _, credential):
        if partition in ["aws", "aws-gov"]:
            super().__init__(partition, ["global"], credential)
        else:
            super().__init__(partition, ["cn-northwest-1", "cn-northeast-1"], credential)

    def build_resource(self, client: any, dist: dict) -> AWSResource:
        arn = dist["ARN"]
        resource_tags = client.list_tags_for_resource(Resource=arn)

        return AWSResource(
            self.category,
            self._build_arn(client, arn),
            [AWSResourceTag(tag["Key"], tag["Value"]) for tag in resource_tags.get("Tags", {}).get("Items", [])],
            dist,
        )

    def _list_resources(self, client: any) -> List[AWSResource]:
        resources = []
        paginator = client.get_paginator("list_distributions").paginate()

        for page in paginator:
            distributions = page.get("DistributionList", {})
            for dist in distributions.get("Items", []):
                resources.append(self.build_resource(client, dist))

        return resources

    @property
    def _service_name(self) -> str:
        return "cloudfront"

    @property
    def _arn_resource_type(self) -> str:
        return "distribution"

    @property
    def category(self) -> str:
        return "CloudFront"



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

import logging
from typing import Dict, List, Tuple

from awstt import output
from awstt.config import Config, ConfigError, Resource, Tag, check_config
from awstt.evals import eval_expression
from awstt.worker.thread import Scanner, ScanningThread, Tagger, TaggingThread
from awstt.worker.types import AWSResource, AWSResourceTag, RegionalTaggingTarget, TaggingResponse
from awstt.worker.utils import filter_tags, is_arn, parse_arn


logger = logging.getLogger(__name__)


def _concat_filter(g_filter: str, r_filter: str) -> str:
    return f"{g_filter} && {r_filter}" if all((g_filter, r_filter)) else g_filter or r_filter


def _concat_tags(g_tags: List[Tag], r_tags: List[Tag]) -> List[Tag]:
    if any(isinstance(t, Tag) for t in g_tags + r_tags):
        tags = {t.key: t.value for t in g_tags} | {t.key: t.value for t in r_tags}
        return [Tag(key=k, value=v) for k, v in tags.items()]
    else:
        return g_tags + r_tags


def _run_list_cmd(config: Config, console: output.Console) -> List[Tuple[Resource, List[AWSResource]]]:
    for name in Scanner.list_available():
        scanner = Scanner.by_name(name)(config.partition, config.regions, config.credential)

        if len(config.resources) == 0:
            ScanningThread.add(
                scanner,
                Resource(target=name, filter=config.filter, tags=config.tags, force=config.force),
            )
        else:
            for res in config.resources:
                if isinstance(res, str):
                    target = res
                else:
                    target = res.target

                if (is_arn(target) and scanner.is_supported_arn(target)) or (name.lower() == target.lower()):
                    if not isinstance(res, str):
                        ScanningThread.add(
                            scanner,
                            Resource(
                                target=target,
                                filter=_concat_filter(config.filter, res.filter),
                                tags=_concat_tags(config.tags, res.tags),
                                force=config.force if res.force is None else res.force,
                            ),
                        )
                    else:
                        ScanningThread.add(
                            scanner,
                            Resource(
                                target=target,
                                filter=config.filter,
                                tags=config.tags,
                                force=config.force,
                            ),
                        )

    return ScanningThread.execute(console, config.env)


def _group_tagging_resources(
    env: dict, resources: List[AWSResource], tags: List[Tag], force: bool
) -> List[RegionalTaggingTarget]:
    class TaggingGroup:
        def __init__(self, rs: List[AWSResource], ts: List[Tag]):
            self.resources = rs
            self.tags = ts

    # group resources by region first then by eval tags

    groups: Dict[str, Dict[str, TaggingGroup]] = {}

    for resource in resources:
        region = parse_arn(resource.arn).region
        if region not in groups:
            groups[region] = {}

        region_group = groups[region]

        real_tags = []
        for tag in tags:
            tag_key = eval_expression(tag.key, env, resource.spec)
            tag_val = eval_expression(tag.value, env, resource.spec)
            if force:
                real_tags.append(Tag(key=tag_key, value=tag_val))
            elif not any(t.key == tag_key for t in resource.tags):
                real_tags.append(tag)

        if len(real_tags) == 0:
            continue

        group_key = " | ".join(map(lambda t: "<~~%s=%s~~>" % (t.key, t.value), real_tags))
        if group_key not in region_group:
            region_group[group_key] = TaggingGroup([], real_tags)

        region_group[group_key].resources.append(resource)

    targets = []
    for region, region_group in groups.items():
        for group in region_group.values():
            if len(group.tags) == 0:
                continue

            targets.append(
                RegionalTaggingTarget(
                    region, group.resources, [AWSResourceTag(key=t.key, value=t.value) for t in group.tags]
                )
            )

    return targets


def _run_set_cmd(
    config: Config, inputs: List[Tuple[Resource, List[AWSResource]]], console: output.Console
) -> List[TaggingResponse]:
    tagger = Tagger(config.partition, config.credential)

    for expect, resources in inputs:
        if len(resources) == 0:
            continue

        g_resources = _group_tagging_resources(config.env, resources, expect.tags, expect.force)
        TaggingThread.add(g_resources)

    return TaggingThread.execute(tagger, console)


def _group_untagging_resources(env: dict, resources: List[AWSResource], tags: List[str]) -> List[RegionalTaggingTarget]:
    resources_with_filtered_tags = filter_tags(resources, tags, env)

    regional_resources: Dict[str, Dict] = {}

    for item in resources_with_filtered_tags:
        resource, tags = item[0], item[1]

        region = parse_arn(resource.arn).region
        if region not in regional_resources:
            regional_resources[region] = {}

        tags_key = ",".join(tags)
        if tags_key not in regional_resources[region]:
            regional_resources[region][tags_key] = {"tags": tags, "resources": []}

        regional_resources[region][tags_key]["resources"].append(resource)

    resp = []
    for region, item in regional_resources.items():
        for tags_key, target in item.items():
            resp.append(RegionalTaggingTarget(region, target["resources"], target["tags"]))

    return resp


def _run_unset_cmd(
    config: Config, inputs: List[Tuple[Resource, List[AWSResource]]], console: output.Console
) -> List[TaggingResponse]:
    tagger = Tagger(config.partition, config.credential, action="unset")
    for expect, resources in inputs:
        if len(resources) == 0:
            continue

        g_resources = _group_untagging_resources(config.env, resources, expect.tags)
        TaggingThread.add(g_resources)

    return TaggingThread.execute(tagger, console)


def run(config: Config):
    echo_config = config.dict()
    console = output.console()

    if config.credential and config.credential.access_key is not None:
        echo_config["credential"]["access_key"] = "******"
    if config.credential and config.credential.secret_key is not None:
        echo_config["credential"]["secret_key"] = "******"

    console.print(
        console.new_panel(
            console.new_pretty(echo_config),
            padding=(1, 1),
            title="[b red]Running with config",
            border_style="bright_blue",
        ),
    )

    logger.info(f"Running with config: \n{echo_config}")

    try:
        check_config(config)
    except ConfigError as e:
        logger.fatal(f"Bad config: {e}")

    resources_list_response = _run_list_cmd(config, console)

    if config.action.lower() == "list":
        categories = {}

        logger.info(f"Tags list executed summary:")

        table = console.new_table(title="Tags List Summary")
        table.add_column("ARN")
        table.add_column("Tags", width=40)
        for _, resources in resources_list_response:
            for res in resources:
                if not categories.get(res.category):
                    categories[res.category] = True
                    logger.info(f"=== {res.category} ===")

                    table.add_section()
                    table.add_row(res.category, "", end_section=True)

                tags = "\n".join(["%s = %s" % (t.key, t.value) for t in res.tags]) if len(res.tags) > 0 else "<No Tags>"
                table.add_row(res.arn, tags)
                logger.info("%-80s\n%s" % (res.arn, tags))

        console.print(table)

    if config.action.lower() in ["set", "unset"]:
        categories = {}

        logger.info(f"Tags {config.action.lower()} executed summary:")

        if config.action.lower() == "set":
            summaries = _run_set_cmd(config, resources_list_response, console)
        else:
            summaries = _run_unset_cmd(config, resources_list_response, console)

        table = console.new_table(title="Tags Set Summary")
        table.add_column("ARN", width=80)
        table.add_column("Status", width=10)
        table.add_column("Hint", width=40)

        for summary in summaries:
            if not categories.get(summary.category):
                categories[summary.category] = True
                logger.info(f"=== {summary.category} ===")

                table.add_section()
                table.add_row(summary.category, "", end_section=True)

            table.add_row(summary.arn, summary.status, summary.hint if summary.hint else "")
            logger.info("%-80s  %-10s  %s" % (summary.arn, summary.status, summary.hint if summary.hint else ""))

        console.print(table)



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

from __future__ import annotations

import logging
import traceback
from abc import ABC, abstractmethod
from typing import List

import boto3

from awstt.config import Credential, Resource
from awstt.worker.registrable import Registrable
from awstt.worker.types import AWSResource
from awstt.worker.utils import detect_region, filter_resources, is_arn, parse_arn


logger = logging.getLogger(__name__)


class Scanner(Registrable, ABC):
    def __init__(self, partition: str, regions: List[str], credential: Credential):
        """
        Initialize a resource scanner to scan resources

        :param partition: The AWS partition of scanner
        :param regions: The AWS regions of scanner
        :param credential: The AWS credential of scanner
        :type credential: Credential
        """

        session = boto3.Session(
            aws_access_key_id=credential.access_key,
            aws_secret_access_key=credential.secret_key,
            profile_name=credential.profile,
        )

        self._session = session
        self._partition = partition
        self._available_regions = regions
        self._account_id = None

    def is_supported_arn(self, arn: str) -> bool:
        info = parse_arn(arn)
        return info.service == self._service_name and info.resource_type == self._arn_resource_type

    def execute(self, *, expect: Resource, env: any) -> List[AWSResource]:
        """
        Scan resources with expect resource config

        :param expect: The target resource config
        :param env: The environment variables
        :return: A list ARNs of resources matched selectors
        :rtype: List[AWSResource]
        """
        logger.debug(f"Starting resources scan - {self.category}")

        # init account_id only when executing, make faster
        self._account_id = self._session.client("sts").get_caller_identity()["Account"]

        # init regions only when executing, make faster
        if len(self._available_regions) == 0:
            self._available_regions = self._session.get_available_regions(
                self._service_name, partition_name=self._partition
            )

        resources: List[AWSResource] = []
        for region in self._available_regions:
            try:
                logger.debug(f"Scanning - {self.category} @ {region}")
                client = self._create_client(self._session, region=detect_region(region, self._partition))
                founded = self._list_resources(client)
                filtered = filter_resources(founded, expect, env)

                resources.extend(filtered)
                logger.debug(
                    f"Finished - {self.category} @ {region}, " f"total - {len(founded)}, matched - {len(filtered)}"
                )
            except Exception as e:
                stack = f"stack:\n{traceback.format_exc()}" if logger.isEnabledFor(logging.DEBUG) else ""
                logger.error(f"Failed - {self.category} @ {region}, error: {e} {stack}")

        logger.debug(f"Finished resources scan - {self.category}, found - {len(resources)}")

        return resources

    def _create_client(self, session: any, *, region: str) -> any:
        """
        A shortcut to create boto3 service client

        :param session: The boto3 session
        :param region: The AWS region of client to connect
        :return: A boto3 service client instance
        """

        # avoid s3 get_bucket_tagging takes too long time if no tags set
        # config = boto3.session.Config(connect_timeout=30, retries={"max_attempts": 0})

        return session.client(self._service_name, region_name=region)

    def _build_arn(self, client: any, arn_or_id: str) -> str:
        """
        A shortcut to build AWS ARN string for most Services

        :param client: The boto3 client used to get the resource, contains partition and region information
        :param arn_or_id: The original string return from `_get_resources`, can be an ARN or resource id
        :return: AWS ARN string
        """
        if is_arn(arn_or_id):
            return arn_or_id

        prefix = (
            f"arn:{self._partition}:{self._arn_service_type}:{detect_region(client.meta.region_name, self._partition)}"
        )

        return f"{prefix}:{self._account_id}:{self._arn_resource_type}/{arn_or_id}"

    @abstractmethod
    def _list_resources(self, client: any) -> List[AWSResource]:
        """
        list resources in region with filters

        :param client: A boto3 client
        :return: A list ARNs of resources matched filters
        :rtype: List[AWSResource]
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def _service_name(self) -> str:
        """
        The service name used to create boto3 client instance

        :return: A boto3 client service name
        """
        raise NotImplementedError

    @property
    def _arn_service_type(self) -> str:
        """
        The ARN service type to detect or build ARN, default is the same as `_service_name`

        :return: An ARN service type string
        """
        return self._service_name

    @property
    @abstractmethod
    def _arn_resource_type(self) -> str:
        """
        The ARN resource type to detect or build ARN

        :return: An ARN resource type string
        """
        return ""

    @property
    @abstractmethod
    def category(self) -> str:
        """
        The name of resource scanner

        :return: Scanner name
        """
        raise NotImplementedError

    @property
    def arn_pattern(self) -> str:
        """
        The ARN pattern of resource scanner

        :return: Scanner ARN pattern
        """
        return f"arn:{self._partition}:{self._arn_service_type}:*:*:{self._arn_resource_type}/*"



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

import datetime
import logging
import math


# add any needed builtins back in
_safe_list = dict()
_safe_list["abs"] = abs
_safe_list["int"] = int
_safe_list["str"] = str
_safe_list["float"] = float
_safe_list["math"] = math
_safe_list["now"] = datetime.datetime.now
_safe_list["date"] = datetime.datetime.date
_safe_list["time"] = datetime.datetime.time
_safe_list["today"] = datetime.datetime.today
_safe_list["timedelta"] = datetime.timedelta

logger = logging.getLogger(__name__)


def eval_expression(exp: str, env: any = None, spec: any = None) -> any:
    data = {"env": env if env is not None else {}, "spec": spec if spec is not None else {}, **_safe_list}

    if exp.lower() in ["type", "class", "def"]:
        return exp

    if any(sub in exp for sub in ["import", "__import__"]):
        logger.fatal(f"Expression contains dangerous code - {exp}")

    # noinspection PyBroadException
    try:
        return eval(exp, data)
    except Exception as e:
        return exp



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

from dataclasses import asdict, dataclass
from typing import Dict, List, Union


@dataclass(frozen=True, init=True, repr=True, order=True)
class _BaseType:
    def dict(self) -> Dict[str, any]:
        return asdict(self)


@dataclass(frozen=True, init=True, repr=True, order=True)
class ArnInfo(_BaseType):
    partition: str
    service: str
    region: str
    account_id: str
    resource_type: str
    resource: str


@dataclass(frozen=True, init=True, repr=True, order=True)
class AWSResourceTag(_BaseType):
    key: str
    value: str


@dataclass(frozen=True, init=True, repr=True, order=True)
class AWSResource(_BaseType):
    category: str
    arn: str
    tags: List[AWSResourceTag]
    spec: dict


@dataclass(frozen=True, init=True, repr=True, order=True)
class ResourceFilter(_BaseType):
    pattern: str
    conditions: List[str]


@dataclass(frozen=True, init=True, repr=True, order=True)
class RegionalTaggingTarget(_BaseType):
    region: str
    resources: List[AWSResource]
    tags: List[Union[AWSResourceTag, str]]


@dataclass(frozen=True, init=True, repr=True, order=True)
class TaggingResponse(_BaseType):
    category: str
    arn: str
    status: str
    hint: str = None



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

import logging
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import List, Tuple

from awstt import output
from awstt.config import Resource
from awstt.worker.scanner import Scanner
from awstt.worker.tagger import Tagger
from awstt.worker.types import (
    AWSResource,
    RegionalTaggingTarget,
    TaggingResponse,
)


logger = logging.getLogger(__name__)

SCANNING_THREAD_LIMIT = 10
RATE_LIMIT_EXCEEDED = "Rate exceeded"


class ScanningThread:
    _workers: List["ScanningThread.Worker"] = []

    class Worker:
        def __init__(self, scanner: Scanner, expect: Resource):
            self.scanner = scanner
            self.expect = expect

        def execute(self, env: any) -> Tuple[Resource, List[AWSResource]]:
            resources = []
            try:
                resources = self.scanner.execute(expect=self.expect, env=env)
            except KeyboardInterrupt:
                logger.warning(f"Scanning process terminated - {self.scanner.category}")
            except Exception as e:
                logger.error(f"Scanning process failed - {self.scanner.category}, error - {e}")
            finally:
                return self.expect, resources

    @classmethod
    def add(cls, scanner: Scanner, resource: Resource):
        cls._workers.append(ScanningThread.Worker(scanner, resource))

    @classmethod
    def execute(cls, console: output.Console, env: any) -> List[Tuple[Resource, List[AWSResource]]]:
        logger.debug(f"Executing resources scan, total - {len(cls._workers)}")

        resources = []
        with ThreadPoolExecutor(max_workers=SCANNING_THREAD_LIMIT) as executor:
            progress = console.new_progress()
            task_id = progress.add_task("Scanning...".ljust(15), total=len(cls._workers))
            progress.start()

            futures = [executor.submit(worker.execute, env) for worker in cls._workers]
            for future in as_completed(futures):
                progress.update(task_id, advance=1)
                resources.append(future.result())

        progress.stop()
        logger.debug(f"Finished resources scan, found - {len(resources)}")

        return resources


class TaggingThread:
    _workers: List["TaggingThread.Worker"] = []

    class Worker:
        def __init__(self, targets: List[RegionalTaggingTarget]):
            self.targets = targets

        def execute(self, tagger: Tagger) -> List[TaggingResponse]:
            t_dict = [target.dict() for target in self.targets]

            responses = []
            try:
                responses = tagger.execute(self.targets)
            except KeyboardInterrupt:
                logger.warning(f"Tags set/unset process terminated, targets - {t_dict}")
            except Exception as e:
                logger.error(f"Tags set/unset process failed, targets - {t_dict}, exception - {e}")
            finally:
                return responses

        @property
        def total_targets(self) -> int:
            total = 0
            for target in self.targets:
                total += len(target.resources)

            return total

    @classmethod
    def add(cls, targets: List[RegionalTaggingTarget]):
        cls._workers.append(TaggingThread.Worker(targets))

    @classmethod
    def execute(cls, tagger: Tagger, console: output.Console) -> List[TaggingResponse]:
        logger.debug(f"Executing resources tagging...")

        progress = console.new_progress()
        progress.add_task("Tagging...".ljust(15), total=None)
        progress.start()

        rate_exceeded_targets = []

        responses = []
        for worker in cls._workers:
            resp = worker.execute(tagger=tagger)
            for r in resp:
                if r.hint != RATE_LIMIT_EXCEEDED:
                    responses.append(r)
                else:
                    for target in worker.targets:
                        for res in target.resources:
                            if res.arn == r.arn:
                                rate_exceeded_targets.append((target.region, res, target.tags))

        while 1:
            if len(rate_exceeded_targets) == 0:
                break

            targets_map = {}
            for item in rate_exceeded_targets:
                region, res, tags = item[0], item[1], item[2]
                tag_str = ",".join([f"{t.key}={t.value}" for t in tags])

                if region not in targets_map:
                    targets_map[region] = {}
                region_map = targets_map[region]
                if tag_str not in region_map:
                    region_map[tag_str] = {"tags": tags, "resources": []}
                region_map[tag_str]["resources"].append(res)

            regional_targets: List[RegionalTaggingTarget] = []
            for region, targets in targets_map.items():
                for _, tag_group in targets.items():
                    regional_targets.append(RegionalTaggingTarget(region, tag_group["resources"], tag_group["tags"]))

            resp = tagger.execute(regional_targets)
            for r in resp:
                if r.hint != RATE_LIMIT_EXCEEDED:
                    responses.append(r)
                    for item in rate_exceeded_targets:
                        if item[1].arn == r.arn:
                            rate_exceeded_targets.remove(item)

            if len(rate_exceeded_targets) > 0:
                time.sleep(0.3)  # avoid throttling

        progress.stop()
        logger.debug(f"Finished batch resources tagging")

        responses.sort(key=lambda x: f"{x.category}--{x.arn}")
        return responses



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

from typing import List

from awstt.worker.scanner import Scanner
from awstt.worker.types import AWSResource, AWSResourceTag


@Scanner.register("EC2:VPC")
class VPCScanner(Scanner):
    def _build_resource(self, client: any, vpc: dict) -> AWSResource:
        return AWSResource(
            self.category,
            self._build_arn(client, vpc["VpcId"]),
            [AWSResourceTag(tag["Key"], tag["Value"]) for tag in vpc.get("Tags", [])],
            vpc,
        )

    def _list_resources(self, client: any) -> List[AWSResource]:
        resources = []
        paginator = client.get_paginator("describe_vpcs").paginate()

        for page in paginator:
            for vpc in page.get("Vpcs", []):
                resources.append(self._build_resource(client, vpc))

        return resources

    @property
    def _service_name(self) -> str:
        return "ec2"

    @property
    def _arn_resource_type(self) -> str:
        return "vpc"

    @property
    def category(self) -> str:
        return "EC2:VPC"



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

from typing import List

from awstt.worker.scanner import Scanner
from awstt.worker.types import AWSResource, AWSResourceTag


@Scanner.register("S3:Bucket")
class S3BucketScanner(Scanner):
    def build_resource(self, client: any, bucket: dict) -> AWSResource:
        # noinspection PyBroadException
        try:
            resource_tags = client.get_bucket_tagging(Bucket=bucket["Name"])
        except:  # noqa: E722
            resource_tags = {"TagSet": []}

        return AWSResource(
            self.category,
            self._build_arn(client, bucket["Name"]),
            [AWSResourceTag(tag["Key"], tag["Value"]) for tag in resource_tags.get("TagSet", [])],
            bucket,
        )

    def _list_resources(self, client: any) -> List[AWSResource]:
        resources = []
        paginator = client.get_paginator("list_buckets").paginate()

        for page in paginator:
            for bucket in page.get("Buckets", []):
                if client.get_bucket_location(Bucket=bucket["Name"])["LocationConstraint"] == client.meta.region_name:
                    resources.append(self.build_resource(client, bucket))

        return resources

    def _build_arn(self, client: any, bucket_name: str) -> str:
        return f"arn:aws:s3:::{bucket_name}"

    @property
    def _service_name(self) -> str:
        return "s3"

    @property
    def _arn_resource_type(self) -> str:
        return ""

    @property
    def category(self) -> str:
        return "S3:Bucket"



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

from typing import List

from awstt.worker.scanner import Scanner
from awstt.worker.types import AWSResource, AWSResourceTag


@Scanner.register("CloudTrail")
class CloudTrailScanner(Scanner):
    def __init__(self, partition, _, credential):
        super().__init__(partition, ["global"], credential)

    def build_resource(self, client: any, trail: dict) -> AWSResource:
        arn = trail["TrailARN"]
        resource_tags = []
        for page in client.get_paginator("list_tags").paginate(ResourceIdList=[arn]):
            for item in page.get("ResourceTagList", []):
                resource_tags.extend(item.get("Items", []))

        return AWSResource(
            self.category,
            self._build_arn(client, arn),
            [AWSResourceTag(tag["Key"], tag["Value"]) for tag in resource_tags],
            trail,
        )

    def _list_resources(self, client: any) -> List[AWSResource]:
        resources = []
        paginator = client.get_paginator("list_trails").paginate()

        for page in paginator:
            for trail in page.get("Trails", []):
                resources.append(self.build_resource(client, trail))

        return resources

    @property
    def _service_name(self) -> str:
        return "cloudtrail"

    @property
    def _arn_resource_type(self) -> str:
        return "trail"

    @property
    def category(self) -> str:
        return "CloudTrail"



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

import re
from datetime import date, datetime, time, timedelta
from json import JSONDecoder, JSONEncoder
from typing import List, Tuple

import jmespath
from dateutil import parser

from awstt.config import Resource
from awstt.evals import eval_expression
from awstt.worker.types import ArnInfo, AWSResource


class DateTimeEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime, date, time)):
            return obj.isoformat()
        elif isinstance(obj, timedelta):
            return (datetime.min + obj).time().isoformat()

        return super(DateTimeEncoder, self).default(obj)


class DateTimeDecoder(JSONDecoder):
    def __init__(self, *args, **kwargs):
        JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    @staticmethod
    def object_hook(d):
        for key, value in d.items():
            try:
                d[key] = parser.parse(value)
            except (ValueError, AttributeError):
                pass
        return d


def is_arn(inputs: str) -> bool:
    return inputs.lower().startswith("arn:")


def is_arn_wild_match(pattern: str, inputs: str) -> bool:
    self = parse_arn(pattern)
    other = parse_arn(inputs)

    matched = (
        self.partition == other.partition
        and self.service == other.service
        and (self.region == other.region or self.region in ["*", ""] or other.region in ["*", ""])
        and (self.account_id == other.account_id or self.account_id in ["*", ""] or other.account_id in ["*", ""])
        and self.resource_type == other.resource_type
        and (self.resource == other.resource or self.resource in ["*", ""] or other.resource in ["*", ""])
    )

    return matched


def parse_arn(inputs: str) -> ArnInfo:
    pattern = (
        r"^arn:(?P<Partition>[^:\n]*)"
        r":(?P<Service>[^:\n]*)"
        r":(?P<Region>[^:\n]*)"
        r":(?P<AccountID>[^:\n]*)"
        r":(?P<Ignore>(?P<ResourceType>[^:\/\n]*)[:\/])?(?P<Resource>.*)$"
    )

    m = re.match(pattern, inputs, re.IGNORECASE)

    return ArnInfo(
        m.group("Partition"),
        m.group("Service"),
        m.group("Region"),
        m.group("AccountID"),
        m.group("ResourceType"),
        m.group("Resource"),
    )


def filter_resources(resources: List[AWSResource], expect: Resource, env: any) -> List[AWSResource]:
    matched_list = []
    for res in resources:
        if is_arn(expect.target) and not is_arn_wild_match(expect.target, res.arn):
            continue

        if not is_arn(expect.target) and expect.target.lower() != res.category.lower():
            continue

        if expect.filter is None or len(expect.filter) == 0:
            matched = True
        else:
            met = jmespath.search(eval_expression(expect.filter, env, res.spec), res.dict())
            if isinstance(met, bool) and met:
                matched = True
            elif isinstance(met, list) and len(met) > 0:
                matched = True
            elif met:
                matched = True
            else:
                matched = False

        if matched:
            if not any(res.arn == m.arn for m in matched_list):
                matched_list.append(res)

    return matched_list


def filter_tags(resources: List[AWSResource], filters: List[str], env: any) -> List[Tuple[AWSResource, List[str]]]:
    if len(filters) == 0:
        return []

    resources_with_filtered_tags = []

    for resource in resources:
        tags: List[str] = []

        for f in filters:
            if f == "*":
                tags.extend([t.key for t in resource.tags])
                continue

            if not f.lower().startswith("tags["):
                f = f"tags[?key=='{f}']"

            ts = jmespath.search(eval_expression(f, env, resource.spec), resource.dict())

            if ts is None:
                continue
            elif isinstance(ts, list) and len(ts) > 0:
                if isinstance(ts[0], str):
                    tags.extend(ts)
                elif isinstance(ts[0], dict) is not None:
                    tags.extend([t["key"] for t in ts if t.get("Key", None) is not None])
            elif isinstance(ts, dict) and ts.get("key", None) is not None:
                tags.append(ts["key"])
            elif isinstance(ts, str):
                tags.append(ts)

        if len(tags) > 0:
            resources_with_filtered_tags.append((resource, tags))

    return resources_with_filtered_tags


def detect_region(region_name: str, partition: str) -> str:
    if region_name == "" or "global" in region_name:
        if partition == "aws":
            return "us-east-1"
        if partition == "aws-gov":
            return "us-gov"

    return region_name



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

if __name__ == "__main__":
    from awstt.cli import run

    run()



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

import logging
import os
import sys
from datetime import datetime
from typing import Any, Callable, Optional, Union


class FakeColumn:
    def __init__(self, _name: str, _width: Optional[int] = None):
        pass


class FakeRow:
    def __init__(self, *values, _end_section: bool = False):
        pass


class FakeTable:
    def __init__(self, _title: str):
        pass

    def add_section(self):
        pass

    def add_column(self, name: str, width: Optional[int] = None):
        pass

    def add_row(self, *values, end_section: bool = False):
        pass


class FakeProgress:
    def __init__(self):
        self.tasks = []

    def add_task(self, description: str, total: float = 100.0):
        self.tasks.append((description, total))
        return len(self.tasks) + 1

    def update(self, task_id: int, advance: float = 1.0):
        pass

    def start(self):
        pass

    def stop(self):
        pass

    def remove_task(self, task_id: int):
        pass


class FakePanel:
    pass


class FakeConsole:
    # noinspection PyMethodMayBeStatic
    def print(self, *objects, **kwargs):
        pass

    def print_json(
        self,
        json: Optional[str] = None,
        *,
        data: Any = None,
        indent: Union[None, int, str] = 2,
        highlight: bool = True,
        skip_keys: bool = False,
        ensure_ascii: bool = False,
        check_circular: bool = True,
        allow_nan: bool = True,
        default: Optional[Callable[[Any], Any]] = None,
        sort_keys: bool = False,
    ) -> None:
        pass

    def pprint(
        self,
        objects: Any,
    ) -> None:
        pass

    # noinspection PyMethodMayBeStatic
    def new_table(self, title: str, **_kwargs):
        return FakeTable(title)

    # noinspection PyMethodMayBeStatic
    def new_progress(self):
        return FakeProgress()

    # noinspection PyMethodMayBeStatic
    def new_panel(self, *_kwarg, **_kwargs):
        return FakePanel()

    # noinspection PyMethodMayBeStatic
    def new_pretty(self, *args, **kwargs):
        pass

    @staticmethod
    def is_fake():
        return True


if os.getenv("LAMBDA_TASK_ROOT") is None:
    from rich.console import Console as RichConsole
    from rich.panel import Panel as RichPanel
    from rich.pretty import Pretty, pprint
    from rich.progress import BarColumn
    from rich.progress import Progress as RichProgress
    from rich.progress import SpinnerColumn, TaskProgressColumn, TextColumn, TimeElapsedColumn
    from rich.table import Table as RichTable

    class Console(RichConsole):
        # noinspection PyMethodMayBeStatic
        def new_table(self, title: str, **kwargs):
            return RichTable(title=title, **kwargs)

        # noinspection PyMethodMayBeStatic
        def new_progress(self):
            return RichProgress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                BarColumn(),
                TaskProgressColumn(),
                TimeElapsedColumn(),
                console=self,
            )

        # noinspection PyMethodMayBeStatic
        def new_panel(self, *args, **kwargs):
            return RichPanel.fit(*args, **kwargs)

        # noinspection PyMethodMayBeStatic
        def new_pretty(self, *args, **kwargs):
            return Pretty(*args, **kwargs)

        def pprint(
            self,
            objects: Any,
        ) -> None:
            pprint(objects, console=self)

        @staticmethod
        def is_fake():
            return False

else:
    Console = FakeConsole

__console = None


def console() -> Console:
    global __console
    if __console is None:
        __console = Console()
    return __console


class ShutdownHandler(logging.Handler):
    def emit(self, record):
        logging.shutdown()
        sys.exit(1)


log_level = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}


def init_logger(level: str = "error", save: bool = False):
    if os.getenv("LAMBDA_TASK_ROOT") is None:
        from rich.highlighter import NullHighlighter
        from rich.logging import RichHandler

        logging.basicConfig(
            handlers=[
                RichHandler(
                    console=console(),
                    omit_repeated_times=False,
                    # log_time_format="[%Y-%m-%d %H:%M:%S]",
                    highlighter=NullHighlighter(),
                ),
            ],
            format="%(message)s",
        )

        if save:
            file_handler = logging.FileHandler(f"awstt-{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.log")
            file_handler.setFormatter(
                logging.Formatter("[%(asctime)s] | %(levelname)-8s | %(name)s:%(lineno)s - %(message)s")
            )
            logging.getLogger().addHandler(file_handler)
    else:
        import boto3

        boto3.set_stream_logger("", logging.INFO)

    logging.getLogger().setLevel(log_level.get(level, logging.INFO))
    logging.getLogger("botocore").setLevel(logging.ERROR)
    logging.getLogger("boto").setLevel(logging.ERROR)
    logging.getLogger("urllib3").setLevel(logging.ERROR)
    logging.getLogger().addHandler(ShutdownHandler(level=logging.CRITICAL))



#  Copyright (c) 2024 AnyIons, All rights reserved.
#  This file is part of aws-tag-tools, released under the MIT license.
#  See the LICENSE file in the project root for full license details.

import logging
import time
from typing import List, Union

import boto3
import botocore.config

from awstt.config import Credential
from awstt.worker.types import AWSResource, AWSResourceTag, RegionalTaggingTarget, TaggingResponse
from awstt.worker.utils import detect_region


logger = logging.getLogger(__name__)

RESOURCES_PER_REQUEST = 20  # The max limited to tag resources per request


class Tagger(object):
    def __init__(self, partition: str, credential: Credential, action="set"):
        """
        Initialize a resource tagger to tag resources

        :param partition: The AWS partition of tagger
        :type partition: str
        :param credential: The AWS credential of tagger
        :type credential: Credential
        :param action: The action of tagger, support "set" and "delete"
        :type action: str
        """

        session = boto3.Session(
            aws_access_key_id=credential.access_key,
            aws_secret_access_key=credential.secret_key,
            profile_name=credential.profile,
        )

        self._session = session
        self._partition = partition
        self._account_id = session.client("sts").get_caller_identity().get("Account")
        self._action = action

    def execute_one(self, region: str, target: AWSResource, tags: List[Union[AWSResourceTag, str]]) -> TaggingResponse:
        """
        Tag one resource use ResourceGroupsTaggingApi

        :param region: The region of resource
        :type region: str
        :param target: The resource to apply tags
        :type target: AWSResource
        :param tags: The tags to apply
        :type tags: List[Union[AWSResourceTag, str]]
        """
        logger.debug(
            f"Executing resource {'tagging' if self._action == 'set' else 'untagging'}: "
            f"region - {detect_region(region, self._partition)}, "
            f"resource - {target}, "
            f"tags - {[t.dict() for t in tags] if self._action == 'set' else tags}"
        )

        # noinspection SpellCheckingInspection
        client = self._session.client(
            "resourcegroupstaggingapi",
            region_name=detect_region(region, self._partition),
            config=botocore.config.Config(
                retries={"max_attempts": 3, "mode": "standard"},
            ),
        )

        if self._action == "set":
            client_resp = client.tag_resources(
                ResourceARNList=[target.arn],
                Tags={t.key: t.value for t in tags},
            )
        else:
            client_resp = client.untag_resources(
                ResourceARNList=[target.arn],
                TagKeys=tags,
            )

        hint = client_resp.get("FailedResourcesMap", {}).get(target.arn, {}).get("ErrorMessage", None)

        return TaggingResponse(
            category=target.category,
            arn=target.arn,
            status="Success" if hint is None else "Failed",
            hint=hint,
        )

    def execute(self, targets: List[RegionalTaggingTarget]) -> List[TaggingResponse]:
        """
        Tag resources use ResourceGroupsTaggingApi

        :param targets: The regional resources to apply tags
        :type targets: List[RegionalTaggingTarget]
        """

        responses = []

        for target in targets:
            logger.debug(
                f"Executing resources {'tagging' if self._action == 'set' else 'untagging'}: "
                f"region - {detect_region(target.region, self._partition)}, "
                f"resources - {[res.arn for res in target.resources]}, "
                f"tags - {[t.dict() for t in target.tags] if self._action == 'set' else target.tags}"
            )

            # noinspection SpellCheckingInspection,DuplicatedCode
            client = self._session.client(
                "resourcegroupstaggingapi",
                region_name=detect_region(target.region, self._partition),
                config=botocore.config.Config(
                    retries={"max_attempts": 3, "mode": "standard"},
                ),
            )

            if len(target.resources) == 0 or len(target.tags) == 0:
                continue

            """limit resources per request"""
            for chunks in [
                target.resources[i : i + RESOURCES_PER_REQUEST]  # noqa: E203
                for i in range(0, len(target.resources), RESOURCES_PER_REQUEST)
            ]:
                if self._action == "set":
                    client_resp = client.tag_resources(
                        ResourceARNList=[c.arn for c in chunks],
                        Tags={t.key: t.value for t in target.tags},
                    )
                else:
                    client_resp = client.untag_resources(
                        ResourceARNList=[c.arn for c in chunks],
                        TagKeys=target.tags,
                    )

                for item in chunks:
                    status = "Success"

                    hint = client_resp.get("FailedResourcesMap", {}).get(item.arn, {}).get("ErrorMessage", None)
                    if hint:
                        status = "Failed"
                    elif self._action == "set":
                        overwrites = []

                        for exists_tag in item.tags:
                            for tag in target.tags:
                                if exists_tag.key == tag.key and exists_tag.value != tag.value:
                                    overwrites.append(f"{tag.key}: {exists_tag.value} -> {tag.value}")

                        if len(overwrites) > 0:
                            status = "Overwrite"
                            hint = ", ".join(overwrites)
                    elif self._action == "unset":
                        original_tag = [t for t in item.tags if t.key in target.tags]
                        if len(original_tag) > 0:
                            hint = ", ".join([f"{t.key}: {t.value} --> <deleted>" for t in original_tag])

                    item_resp = TaggingResponse(
                        category=item.category,
                        arn=item.arn,
                        status=status,
                        hint=hint,
                    )
                    responses.append(item_resp)

                time.sleep(0.3)  # avoid throttling

            failures = len([res for res in responses if res.status == "Failed"])
            logger.debug(
                f"Finished resources {'tagging' if self._action == 'set' else 'untagging'}: "
                f"region - {target.region}, "
                f"resources - {[res.arn for res in target.resources]}, "
                f"tags - {[t.dict() for t in target.tags] if self._action == 'set' else target.tags}",
                f"successes - {len(target.resources) - failures}, " f"failures - {failures}",
            )

        return responses
